;; This file contains a DrScheme library version of the functional graphics
;; system for use with Concrete Abstractions: An Introduction to Computer
;; Science Using Scheme, by Max Hailperin, Barbara Kaiser, and Karl Knight.
;;
;; This version (tested w/DrScheme v370), unlike the previously distributed
;; version, represents images directly as DrScheme's image snips, which has the
;; virtue that "Insert Image ..." can be used to bring in an image file (e.g.
;; a photo) and then use it for things like pinwheeling.
;;
;; At the moment, color images are treated as opaque, whereas monochrome images
;; are treated as transparent (the white regions are transparent, that is).  This
;; is relevant for the overlay operation.  If the first argument to overlay is a
;; color image, it will totally obscure the second argument.  However, if the
;; first argument to overlay is a monochrome image (including those produced with
;; filled-triangle and line, or by operating on monochrome images), then it will
;; overlay nicely onto the second argument, regardless of whether the second is
;; monochrome or color.
;;
;; To get around this limitation of color (or grayscale) images, they can be
;; converted to monochrome using a newly added operation, threshold.  For
;; example, if photo is a color image, then
;;  (threshold photo 200)
;; is a monochrome image formed by translating all pixels where the sum of
;; the red, green, and blue components (each on a scale of 0 to 255) exceeds
;; 200 into white and the rest of the pixels into black.  Choice of a suitable
;; threshold is left to experimentation.
;;
;; The invert operation can be applied even to color images, though the result
;; isn't all that useful.

(module fungraph mzscheme
  (provide get-default-image-size set-default-image-size!
           line quarter-turn-right mirror-image invert threshold
           overlay resize-image stack filled-triangle write-image-as-epsf
	   test-bb nova-bb bitw-bb rcross-bb corner-bb
	   zero-bb one-bb two-bb three-bb four-bb
	   five-bb six-bb seven-bb eight-bb nine-bb)
  
  (require (lib "class.ss"))
  (require (lib "mred.ss" "mred"))
  
  (define default-image-size 100)
  
  (define get-default-image-size
    (lambda ()
      default-image-size))
  
  (define set-default-image-size!
    (lambda (new)
      (set! default-image-size new)))
  
  (define mirror-image
    (lambda (i)
      (let* ((bm (send i get-bitmap))
             (w (send bm get-width))
             (h (send bm get-height))
             (sz (* w h 4))
             (bs (make-bytes sz))
             (bs2 (make-bytes sz)) ; could reuse bs
             (bm2 (make-object bitmap% w h (not (send bm is-color?))))
             (dc (new bitmap-dc% (bitmap bm2))))
        (send bm get-argb-pixels 0 0 w h bs)
        (let loop1 ((y 0))
          (if (= y h)
              (begin (send dc set-argb-pixels 0 0 w h bs2)
                     (send dc set-bitmap #f)
                     (make-object image-snip% bm2))
              (let loop2 ((x 0))
                (if (= x w)
                    (loop1 (+ y 1))
                    (let ((x2 (- w x 1)))
                      (let loop3 ((c 0))
                        (if (= c 4)
                            (loop2 (+ x 1))
                            (let ((pos (+ c (* 4 (+ x (* y w)))))
                                  (pos2 (+ c (* 4 (+ x2 (* y w))))))
                              (bytes-set! bs2 pos2 (bytes-ref bs pos))
                              (loop3 (+ c 1)))))))))))))
  (define invert
    (lambda (i)
      (let* ((bm (send i get-bitmap))
             (w (send bm get-width))
             (h (send bm get-height))
             (sz (* w h 4))
             (bs (make-bytes sz))
             (bm2 (make-object bitmap% w h (not (send bm is-color?))))
             (dc (new bitmap-dc% (bitmap bm2))))
        (send bm get-argb-pixels 0 0 w h bs)
        (let loop1 ((y 0))
          (if (= y h)
              (begin (send dc set-argb-pixels 0 0 w h bs)
                     (send dc set-bitmap #f)
                     (make-object image-snip% bm2))
              (let loop2 ((x 0))
                (if (= x w)
                    (loop1 (+ y 1))
                    (let loop3 ((c 1))
                      (if (= c 4)
                          (loop2 (+ x 1))
                          (let ((pos (+ c (* 4 (+ x (* y w))))))
                            (bytes-set! bs pos (- 255 (bytes-ref bs pos)))
                            (loop3 (+ c 1))))))))))))
  
  (define threshold
    (lambda (i t)
      (let* ((bm (send i get-bitmap))
             (w (send bm get-width))
             (h (send bm get-height))
             (sz (* w h 4))
             (bs (make-bytes sz))
             (bm2 (make-object bitmap% w h #t))
             (dc (new bitmap-dc% (bitmap bm2))))
        (send bm get-argb-pixels 0 0 w h bs)
        (let loop1 ((y 0))
          (if (= y h)
              (begin (send dc set-argb-pixels 0 0 w h bs)
                     (send dc set-bitmap #f)
                     (make-object image-snip% bm2))
              (let loop2 ((x 0))
                (if (= x w)
                    (loop1 (+ y 1))
                    (let* ((p (lambda (c)
                                (+ c (* 4 (+ x (* y w))))))
                           (new (if (> (+ (bytes-ref bs (p 1))
                                          (bytes-ref bs (p 2))
                                          (bytes-ref bs (p 3)))
                                       t)
                                    255
                                    0)))
                      (let loop3 ((c 1))
                        (if (= c 4)
                            (loop2 (+ x 1))
                            (begin
                              (bytes-set! bs (p c) new)
                              (loop3 (+ c 1)))))))))))))
  
  (define quarter-turn-right
    (lambda (i)
      (let* ((bm (send i get-bitmap))
             (w (send bm get-width))
             (h (send bm get-height))
             (sz (* w h 4))
             (bs (make-bytes sz))
             (bs2 (make-bytes sz))
             (bm2 (make-object bitmap% h w (not (send bm is-color?))))
             (dc (new bitmap-dc% (bitmap bm2))))
        (send bm get-argb-pixels 0 0 w h bs)
        (let loop1 ((y 0))
          (if (= y h)
              (begin (send dc set-argb-pixels 0 0 h w bs2)
                     (send dc set-bitmap #f)
                     (make-object image-snip% bm2))
              (let loop2 ((x 0))
                (if (= x w)
                    (loop1 (+ y 1))
                    (let loop3 ((c 0))
                      (if (= c 4)
                          (loop2 (+ x 1))
                          (let ((pos (+ c (* 4 (+ x (* y w)))))
                                (pos2 (+ c (* 4 (+ (- h y 1) (* x h))))))
                            (bytes-set! bs2 pos2 (bytes-ref bs pos))
                            (loop3 (+ c 1))))))))))))
  
  (define stack
    (lambda (top bottom)
      (let* ((top-bm (send top get-bitmap))
             (bottom-bm (send bottom get-bitmap))
             (top-w (send top-bm get-width))
             (top-h (send top-bm get-height))
             (w (if (= (send bottom-bm get-width) top-w)
                    top-w
                    (error "can't stack images of different widths" top bottom)))
             (bottom-h (send bottom-bm get-height))
             (h (+ top-h bottom-h))
             (bm (make-object bitmap% w h (not (or (send top-bm is-color?)
                                                   (send bottom-bm is-color?)))))
             (dc (new bitmap-dc% (bitmap bm))))
        (send dc clear)
        (send dc draw-bitmap bottom-bm 0 top-h)
        (send dc draw-bitmap top-bm 0 0)
        (send dc set-bitmap #f)
        (make-object image-snip% bm))))
  
  (define overlay
    (lambda (top bottom)
      (let* ((top-bm (send top get-bitmap))
             (bottom-bm (send bottom get-bitmap))
             (top-w (send top-bm get-width))
             (top-h (send top-bm get-height))
             (w (if (= (send bottom-bm get-width) top-w)
                    top-w
                    (error "can't overlay images of different widths" top bottom)))
             (bottom-h (send bottom-bm get-height))
             (h (if (= top-h bottom-h)
                    top-h
                    (error "can't overlay images of different heights" top bottom)))
             (bm (make-object bitmap% w h (not (or (send top-bm is-color?)
                                                   (send bottom-bm is-color?)))))
             (dc (new bitmap-dc% (bitmap bm))))
        (send dc clear)
        (send dc draw-bitmap bottom-bm 0 0)
        (send dc draw-bitmap top-bm 0 0)
        (send dc set-bitmap #f)
        (make-object image-snip% bm))))
  
  (define filled-triangle
    (lambda (x0 y0 x1 y1 x2 y2)
      (let* ((w default-image-size)
             (h default-image-size)
             (pt (lambda (x y)
                   (make-object point%
                     (* (+ x 1) w .5)
                     (* (- 1 y) h .5))))
             (bm (make-object bitmap% w h #t))
             (dc (new bitmap-dc% (bitmap bm))))
        (send dc clear)
        (send dc set-brush (make-object color% 0 0 0) 'solid)
        (send dc draw-polygon (list (pt x0 y0) (pt x1 y1) (pt x2 y2)))
        (send dc set-bitmap #f)
        (make-object image-snip% bm))))
  
  (define line
    (lambda (x0 y0 x1 y1)
      (let* ((w default-image-size)
             (h default-image-size)
             (tx (lambda (x)
                   (* (+ x 1) w .5)))
             (ty (lambda (y)
                   (* (- 1 y) h .5)))
             (bm (make-object bitmap% w h #t))
             (dc (new bitmap-dc% (bitmap bm))))
        (send dc clear)
        (send dc draw-line (tx x0) (ty y0) (tx x1) (ty y1))
        (send dc set-bitmap #f)
        (make-object image-snip% bm))))
  
  (define (resize-image image . wh)
    (let ((width default-image-size)
          (height default-image-size))
      (if (not (null? wh))
          (begin (set! width (car wh))
                 (if (not (null? (cdr wh)))
                     (begin (set! height (cadr wh))
                            (if (not (null? (cddr wh)))
                                (error "too many argument to resize-image")))
                     (set! height width))))
      (let* ((old-bm (send image get-bitmap))
             (old-w (send old-bm get-width))
             (old-h (send old-bm get-height))
             (bm (make-object bitmap% width height (not (send old-bm is-color?))))
             (dc (new bitmap-dc% (bitmap bm))))
        (send dc clear)
        (send dc draw-bitmap-section-smooth old-bm 0 0 width height 0 0 old-w old-h)
        (send dc set-bitmap #f)
        (make-object image-snip% bm))))
  
  (define write-image-as-epsf
    (let ((margin 72.0))
      (lambda (snip filename)
        (let ((pss (current-ps-setup)))
          (send pss set-file filename)
          (send pss set-mode 'file)
          (current-ps-setup pss))
        (let ((dc (make-object post-script-dc% #f)))
          (if (send dc ok?)
              (begin
                (send dc start-doc filename)
                (send dc start-page)
                (let ((wbox (box 0.0))
                      (hbox (box 0.0)))
                  (send snip get-extent dc margin margin wbox hbox #f #f #f #f)
                  (send snip draw dc margin margin margin margin
                        (+ margin (unbox wbox))
                        (+ margin (unbox hbox))
                        0.0 0.0
                        'no-caret))
                (send dc end-page)
                (send dc end-doc)))))))


  ;; The following definitions create the basic quilt-cover block images
  ;; for use with Concrete Abstractions: An Introduction to Computer
  ;; Science Using Scheme by Max Hailperin, Barbara Kaiser, and Karl Knight.
  ;;
  ;; The images defined are:
  ;;  From chapter 1: rcross-bb, corner-bb, test-bb, and nova-bb
  ;;  From chapter 2: bitw-bb
  ;;  Additional images: zero-bb, one-bb, two-bb, three-bb, four-bb,
  ;;                     five-bb, six-bb, seven-bb, eight-bb, nine-bb

  ;; A simple test image, to illustrate transformations.
  (define test-bb
    (filled-triangle 0 1 0 -1 1 -1))

  ;; This one only has two triangles, but makes an interesting pinwheel.
  (define nova-bb
    (overlay (filled-triangle 0 1 0 0 -1/2 0)
	     (filled-triangle 0 0 0 1/2 1 0)))

  ;; Basic block for "Blowing in the Wind" quilting pattern from
  ;; "Quick-and-Easy Strip Quilting" by Helen Whitson Rose, Dover
  ;; Publications, New York, 1989, p. 59.
  (define bitw-bb
    (overlay (overlay (filled-triangle -1 1 0 1 -1/2 1/2)
		      (filled-triangle -1 -1 0 -1 -1 0))
	     (overlay (filled-triangle 1 1 1 0 0 0)
		      (filled-triangle 0 0 1 0 1/2 -1/2))))

  ;; The final two basic blocks defined in this file, rcross-bb and corner-bb,
  ;; are defined in a way intended to be unreasonably hard to understand, because
  ;; defining them is one of the exercises in the text.  It would be easier
  ;; for you to come up with your own definitions from scratch than by puzzling
  ;; these definitons out.  The only point of having them is to let you use the
  ;; blocks without first doing the definitions.  You might as well stop reading
  ;; here, the below is not meant to be readable.
  
  ;; Basic block for "Repeating Crosses" quilting pattern from
  ;; "Quick-and-Easy Strip Quilting" by Helen Whitson Rose, Dover
  ;; Publications, New York, 1989, p. 60.
  (define rcross-bb #f)
  
  ;; A much simpler basic block, with one corner black.
  (define corner-bb #f)
  
  (let ((omb
	 (lambda x
	   (let l
	       ((x (cdr x))
		(y (list-tail (cdr x) (quotient (length x) 2)))
		(z #f)
		(w (car x)))
	     (if (null? y)
		 z
		 (l (cddddr x) (cddddr y)
		    (let* ((v (lambda (v) (/ v w)))
			   (v (filled-triangle (v (car y)) (v (car x))
					       (v (cadr y)) (v (cadr x))
					       (v (caddr y)) (v (caddr x)))))
		      (if z (overlay z v) v)) w))))))
    (set! rcross-bb (omb 2 2 2 1 1 1 2 1 2 1 1 -1 2 -1 -2 1 
			 2 1 -1 -1 2 -2 2
			 -1 1 -1 2 2 1 2 1 1 -1 1 2 2 1 -1 -1 1 1))
    (set! corner-bb (omb -1 -1 -1 0 0 0 -1 -1 -1))
    )
  
  ;; digit images for Exercise 2.23 on page 46
  
  (define zero-bb 'undefined)
  (define one-bb 'undefined)
  (define two-bb 'undefined)
  (define three-bb 'undefined)
  (define four-bb 'undefined)
  (define five-bb 'undefined)
  (define six-bb 'undefined)
  (define seven-bb 'undefined)
  (define eight-bb 'undefined)
  (define nine-bb 'undefined)
  
  (define digit-bb-parameters
    (let ((hoff 'undefined)
	  (voff 'undefined)
	  (thick 'undefined))
      (lambda args
	(cond
	 ((= (length args) 0) (list hoff voff thick))
	 ((not (= (length args) 3))
	  (error 'digit-bb-parameters "expected 3 parameters"))
	 (else (set! hoff (car args))
	       (set! voff (cadr args))
	       (set! thick (caddr args))
	       (let* ((vlen (- 1 voff))
		      (hlen (- 2 (* 2 hoff)))
					;(vlen (/ (- 2 (+ (* 3 thick) (* 2 voff))) 2))
					;(hlen (- 2 (+ (* 2 thick) (* 2 hoff))))
		      (vlen/2 (/ vlen 2))
		      (hlen/2 (/ hlen 2))
		      (thick/2 (/ thick 2)))
		 (letrec
		     ((seg (lambda (x y w h)
			     (let ((x1 (- x (/ w 2)))
				   (x2 (+ x (/ w 2)))
				   (y1 (- y (/ h 2)))
				   (y2 (+ y (/ h 2))))
			       (overlay
				(filled-triangle x1 y1 x1 y2 x2 y2)
				(filled-triangle x2 y2 x2 y1 x1 y1)))))
		      (vseg (lambda (x y) (seg x y thick vlen)))
		      (hseg (lambda (x y) (seg x y hlen thick)))
		      (segs (lambda s
			      (if (null? (cdr s))
				  (car s)
				  (overlay (car s) (apply segs (cdr s)))))))
		   (let ((tc (hseg 0 (+ vlen thick)))
			 (cc (hseg 0 0))
			 (bc (hseg 0 (- (+ vlen thick))))
			 (tl (vseg (- (+ hlen/2 thick/2)) (+ vlen/2 thick/2)))
			 (bl (vseg (- (+ hlen/2 thick/2)) (- (+ vlen/2 thick/2))))
			 (tr (vseg (+ hlen/2 thick/2) (+ vlen/2 thick/2)))
			 (br (vseg (+ hlen/2 thick/2) (- (+ vlen/2 thick/2)))))
		     (set! zero-bb (segs tl tc tr br bc bl))
		     (set! one-bb (segs tr br))
		     (set! two-bb (segs tc tr cc bl bc))
		     (set! three-bb (segs tc tr cc br bc))
		     (set! four-bb (segs tl cc tr br))
		     (set! five-bb (segs tc tl cc br bc))
		     (set! six-bb (segs tc tl cc bl bc br))
		     (set! seven-bb (segs tc tr br))
		     (set! eight-bb (segs tl tc tr cc bl bc br))
		     (set! nine-bb (segs tl tc tr cc bc br))))))))))
  
  ;; horizontal offset, vertical offset, line thickness
  (digit-bb-parameters 1/2 1/3 1/10)
  
  )
